/*
 * Line editor.
 */

#include <termio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/dir.h>
#include <errno.h>
#include "defs.h"

#define MAXIPLINE	512
#define MAXNLAST	512	/* Maximum number of last lines */
#define DNLASTLINES	40	/* Default number of last lines */
#define uint unsigned int
#define ushort unsigned short
#define uchar unsigned char
#define CTL(x)	(x - 0x40)
#define EOFCHAR	CTL('D')

/* Editor externals */
char *newbuf;			/* Pointer to new string */
char *oldbuf;			/* Pointer to last displayed string */
uint cursor;			/* Where the cursor is in the string */
uint cursscrnpos;		/* Where the cursor is on the screen */
uint oldlen;			/* Length of previously displayed string */
uint len;
uint nlastlines;		/* Number of last lines for recall */
char **lastlines;		/* Most recent commands */

int ttyset = 0;

char bbb128;
int bbbindex;

int curinfd;
extern int errno;

struct termio _termio, _oldtermio;

char *roomfor();

xread(fd, buf, nb)
char *buf;
{
	static char started = 0;
	int ret;

	if (!isatty(fd) || (nb == 1))
	{
		ret = READ(fd, buf, nb);
		if (ret < 0)
		{
			prs_buff("Read failed.  Errno=");
			prn_buff(errno);
			prs_buff(", shell says Goodnight.\n");
			flushb();
			EXIT(1);
		}
		return ret;
	}
	curinfd = fd;
	if (!started)
	{
		ledit_init();
		started = 1;
	}
	*buf = 0;
	termstart();
	bbbindex = 0;
	ret = NLEDIT(buf, nb - 1);
	bbbflush();
	termend();
	if (ret)
	{
		strcat(buf, "\n");
		return strlen(buf);
	}
	else
		return 0;
}

termstart()
{
	ioctl(1, TCSBRK, 1);	/* Wait for op to drain before mode change */
	ioctl(1, TCXONC, 1);	/* Restart suspended op, coz ^Q turned off */
	tst(1, ioctl(curinfd, TCGETA, &_termio));
	_oldtermio = _termio;
	ttyset = 1;
	_termio.c_iflag &= ~IXON;
	_termio.c_lflag &= ~ECHO & ~ECHOE & ~ECHONL & ~ICANON;
	_termio.c_ccVEOF = 1;	/* Min. no. of chars */
	tst(2, ioctl(curinfd, TCSETA, &_termio));
}

termend()
{
	tst(3, ioctl(curinfd, TCSETA, &_oldtermio));
	ttyset = 0;
}

tst(n, ret)
{
	extern int errno;
	int errnsave;

	if (ret < 0)
	{
		errnsave = errno;
		prstr(2, "termio failed\n");
		errno = errnsave;
		perror("");
	}
}

/* Initialise the line editor stuff */
ledit_init()
{
	register int i;
	register struct namnod *namnod;

	namnod = (struct namnod *)findnam("NUMLAST");
	
	if (!namnod || !(nlastlines = atoi(namnod->namval)) ||
						(nlastlines > MAXNLAST))
		nlastlines = DNLASTLINES;

	lastlines = (char **)malloc(sizeof(*lastlines) * nlastlines);
	if (resth(0) < 0)
	{
		for (i = 0; i < nlastlines; i++)
		{
			lastlinesi = (char *)malloc(2);
			*lastlinesi = '\0';
		}
	}
}

/*
 * System call 86: Old line editor. Length = MAXIPLINE
 */

ledit(lbuf)
char *lbuf;
{
	termstart();
	NLEDIT(lbuf, MAXIPLINE);
	termend();
}

/*
 * System call 84: Edit the line at *lbuf. Return pointer to buffer.
 * Max line length is passed
 */

nledit(lbuf, length)
register char *lbuf;
int length;
{
	register int ch;
	register int i, igneof = 0;
	register uint llindex;	/* Index into lastlines */
	char *ob;		/* Copy of previously displayed string */
	char arrow;
	int nxtch = 0;
	char *oldlbuf;

	if (length >= MAXIPLINE)
		length = MAXIPLINE-1;
	ob = (char *)malloc(MAXIPLINE);
	oldlbuf = (char *)malloc(MAXIPLINE);
	*oldlbuf = '\0';

	llindex = nlastlines - 1;
	igneof = (long)findnam("IGNOREEOF") != 0;
	newbuf = lbuf;
	oldbuf = ob;
	len = strlen(lbuf);
	cursor = 0;
	oldlen = 0;
	cursscrnpos = 0;

	redisp();
leditloop:
	if (nxtch)
	{
		ch = nxtch;
		nxtch = 0;
	}
	else
	{
		if ((ch = rawgetchar()) < 0)
			ch = EOFCHAR;	/* Force end of input */
	}
	if (!igneof && ch == EOFCHAR && len == 0)
		goto inp_eof;	/* EOFCHAR first entered */
	if ((ch >= ' ') || strchr("WEFGTVXYP", ch + 0x40))
	{	/* Will alter */
		strcpy(oldlbuf, lbuf);
	}
	if (ch == CTL('U'))
	{
		strcpy(lbuf, oldlbuf);
		getlen();
		if (cursor > len)
			cursor = len;
		redisp();
		goto leditloop;
	}

	switch (ch)
	{
	case CTL('A'):	/* ^A Backward word */
		if (cursor)
		{
			do
				cursor--;
			while (iswhite(lbufcursor) && cursor);

			while (cursor && (!iswhite(lbufcursor-1)))
				cursor--;
		}
		break;
	case CTL('B'):	/* ^B Goto end */
		cursor = cursor ? 0 : len;
		break;
	case CTL('D'):	/* ^D Forward char */
		if (cursor < len)
			cursor++;
		break;
	case '\033':	/* ESC - search back for history */
		{
			int i, olen;
			olen = len;
			for (i = nlastlines - 1; i > 0; i--)
			{
				if (!strncmp(lbuf, lastlinesi, olen))
				{
					int ch2;
					strcpy(lbuf, lastlinesi);
					getlen();
					cursor = len;
					redisp();
					ch2 = rawgetchar();
					if (ch2 != '\033')
					{
						nxtch = ch2;
						goto got;
					}
				}
			}
		}
got:
		break;
#ifdef notdef
	case CTL('K'):		/* Filename completion */
		{
			char dirnameMAXIPLINE;
			char filenameMAXIPLINE;
			char *cp, *ocp;
			cp = lbuf + cursor;
			while (cp > lbuf && !iswhite(*cp))
				cp--;
			if (iswhite(*cp))
				cp++;
			ocp = dirname;
			while (*cp && !iswhite(*cp))
				*ocp++ = *cp++;
			*ocp = '\0';
#endif
	case CTL('O'): /* Cycle back through history argument at a time */
		{
			char *getword();
			int olen, i, j, w, k;
			char *word;

			olen = cursor;
			for( i = nlastlines - 1; i > 0; i-- )
			{
				/* Split line up into words */
				w = countwords( lastlinesi );
				for( j = w; j > 0; j-- )
				{
					int ch2, l = 0;
					word = getword( lastlinesi, j );

					while( ch = wordl++ )
						insch( ch, length );

					getlen();
					redisp();
					ch2 = rawgetchar();
					if( ch2 != CTL('O') )
					{
						nxtch = ch2;
						goto ggot;
					}
					cursor = olen;
					for( k = 0; k < strlen( word ); k++ )
						remch();
				}
			}
		}
ggot:
		break;

	case CTL('W'):	/* ^W */
		arrow = 2;
		goto udarrow;
	case CTL('E'):	/* ^E Previous line */
		arrow = 1;		/* Flag: up arrow */
		goto udarrow;
	case CTL('F'):	/* ^F Forward word */
		if (cursor != len)
		{
			do
				cursor++;
			while (!iswhite(lbufcursor) && (cursor != len));
			if (cursor != len)
			{
				do
					cursor++;
				while (iswhite(lbufcursor) && (cursor != len));
			}
		}
		break;
	case CTL('G'):	/* ^G, DEL Delete char forward */
		remch();
		getlen();
		break;
	case 0x7f:
	case '\b':	/* ^H, Delete char backward */
		if (cursor)
		{
			cursor--;
			remch();
			len--;
		}
		break;
	case CTL('S'):	/* ^S Backward char */
		if (cursor)
		{
			cursor--;
			redisp();
		}
		break;
	case CTL('T'):	/* ^T Delete word forward */
		if (!iswhite(lbufcursor))
			while ((lbufcursor) && (!iswhite(lbufcursor)))
				remch();
		else
			while (iswhite(lbufcursor))
				remch();
		getlen();
		break;
	case CTL('V'):	/* ^V Delete line backward */
		strcpy(lbuf, lbuf + cursor);
		cursor = 0;
		getlen();
		break;
	case CTL('X'):	/* ^X Previous line recall */
		arrow = 0;
/*
 * If arrow == 0, scroll forward.
 * If arrow == 1, scroll to previous.
 * If arrow == 2, insert previous line.
 */
udarrow:	if (arrow != 2)
			lbufcursor = len = 0 = '\0';		/* Joke */
		if ((!arrow) && (++llindex >= nlastlines))
			llindex = 0;		/* Wrap */
		i = 0;
		while (ch = lastlinesllindexi++)
			insch(ch, length);
		if ((arrow) && (--llindex >= nlastlines))
			llindex = nlastlines - 1;
		break;
	case CTL('Y'):	/* ^Y Delete line forward */
		lbufcursor = '\0';
		len = cursor;
		break;
	case '\r':	/* Terminator */
	case '\n':
inp_eof:
		cursor = len;
		redisp();	/* Move cursor to end */
		/* Don't copy empty or same string */
		if (len && strcmp(lbuf, lastlinesnlastlines - 1))
		{
			free(lastlines0);
			movmem(&lastlines1, lastlines,
				(nlastlines - 1) * sizeof(*lastlines));
			lastlinesnlastlines - 1 = roomfor(lbuf);
		}
		rawputchar('\n');
		free(ob);
		free(oldlbuf);
		return ((uint)ch == EOFCHAR) ? 0 : (int)lbuf;
	case CTL('P'):	/* ^P Insert control char */
		ch = rawgetchar();
	default:	/* Normal char */
		insch(ch, length);
		break;
	}
	redisp();
	goto leditloop;
}

/*
 * Redisplay buffer
 */

redisp()
{
	register ushort cp;		/* Current buffer position */
	register ushort newspos;	/* Current screen position */
	register ushort newcurspos;	/* New cursor screen position */
	unsigned olsave;		/* Temp save for oldlen */
	register char pe;		/* Print enable */
	register uchar ch;		/* Current char */

	newspos = 0;
	cp = 0;
	pe = 0;

/* Scan the whole new buffer */
	do
	{
		ch = newbufcp;

		if (newspos == cursscrnpos)
			pe = 1;	/* Start printing when the cursor is picked up */

		if (!pe && (ch != oldbufcp))
		{	/* Move the cursor back to where they differ */
			while (cursscrnpos > newspos)
			{
				rawputchar('\b');
				--cursscrnpos;
			}
			pe = 1;
		}

		if (cp == cursor)
			newcurspos = newspos;	/* New cursor position */

		if (ch == '\t')
		{
			do
				if (pe)
					rawputchar(' ');
			while (++newspos & 7);
		}
		else
		{
			if (ch)
			{
				if (ch < 0x20)
				{	/* Control char */
					if (pe)
					{
						rawputchar('^');
						rawputchar(ch | 0x40);
					}
					newspos += 2;
				}
				else
				{
					if (pe)
						rawputchar(ch);
					newspos++;
				}
			}
		}

		oldbufcp = ch;
		cp++;
	}
	while (ch);

	olsave = newspos;

/* New string is out. Print any necessary spaces */
	while (newspos < oldlen)
	{
		rawputchar(' ');
		newspos++;
	}

/* Back up to correct cursor position */
	while (newspos > newcurspos)
	{
		rawputchar('\b');
		--newspos;
	}

	oldlen = olsave;
	cursscrnpos = newspos;
}

/*
 * Remove the char at the cursor
 */

remch()
{
	if (newbufcursor)
		strcpy(newbuf + cursor, newbuf + cursor + 1);
}

/*
 * Insert a char in the buffer
 */

insch(ch, length)
char ch;
{
	register char *ptr;

	if (len < (length - 1))
	{
		ptr = newbuf + cursor++;
		movmem(ptr, ptr + 1, strlen(ptr) + 1);
		*ptr = ch;
		len++;
	}
}

/*
 * Restore len
 */

getlen()
{
	len = strlen(newbuf);
}

char *
roomfor(s)
char *s;
{
	char *ret;

	ret = (char *)malloc(strlen(s) + 2);
	strcpy(ret, s);
	return ret;
}

movmem(from, to, nb)
register char *to, *from;
register int nb;
{
	if (to < from)
		while (nb--)
			*to++ = *from++;
	else
	{
		to += nb;
		from += nb;
		while (nb--)
			*(--to) = *(--from);
	}
}

iswhite(ch)
char ch;
{
	return ((ch == ' ') || (ch == '\t') || (ch == '/'));
}

rawgetchar()
{
	char c;
	int ret;

	bbbflush();
	ret = READ(curinfd, &c, 1);
	if (ret < 0)
	{
		prs_buff("Read failed.  Errno=");
		prn_buff(errno);
		prs_buff(", shell departing\n");
		flushb();
		EXIT(1);
	}
	if (!c)		/* CTL-SPACE! */
		c = ' ';
	return c;
}

rawputchar(c)
char c;
{
	bbbbbbindex++ = c;
	if (bbbindex >= sizeof(bbb))
		bbbflush();
}

cookputs(s)
char *s;
{
	while (*s)
	{
		if (*s < ' ')
		{
			rawputchar('^');
			rawputchar(*(s++) + '@');
		}
		else
			rawputchar(*s++);
	}
}

bbbflush()
{
	if (bbbindex)
	{
		WRITE(1, bbb, bbbindex);
		bbbindex = 0;
	}
}

prstr(fd, s)
char *s;
{
	WRITE(fd, s, strlen(s));
}

prnl()
{
	rawputchar('\n');
}

dohistory()
{
	register int i, len, lastlen;
	register char *cp;
	char in = 0;

	lastlen = 0;
	bbbindex = 0;
	for (i = 0; i < nlastlines; i++)
	{
		cp = lastlinesi;
		if (*cp)
			in = 1;
		if (in)
		{
			len = strlen(cp);
			if (lastlen)
			{
				if (lastlen >= 38 || (len > 38))
				{
					prnl();
					cookputs(cp);
					lastlen = len;
				}
				else
				{
					while (lastlen < 39)
					{
						rawputchar(' ');
						lastlen++;
					}
					cookputs("| ");
					cookputs(cp);
					prnl();
					lastlen = 0;
				}
			}
			else
			{
				cookputs(cp);
				lastlen = len;
			}
		}
	}
	if (lastlen)
		prnl();
	bbbflush();
}

openlastlines(path, name, mode, mode2)
register char *path, *name;
{
	register int	f;

	do
	{
		path = catpath(path, name);
	} while ((f = OPEN(curstak(), mode, mode2)) < 0 && path);
	return(f);
}


/* Save history */
saveh(name)
char *name;
{
	register int fd, i;

	if (!name || !*name)
		name = lastln;
	if ((fd = openlastlines(homenod.namval, name,
											O_WRONLY|O_CREAT|O_TRUNC, 0660))<0)
		return -1;
	
	for (i = nlastlines - 1; i >= 0; i--)
		if (*lastlinesi)
		{
			write (fd, lastlinesi, strlen(lastlinesi));
			write (fd, "\n", 1);
		}
	CLOSE(fd);
	return 0;
}

/* Restore history */
resth(name)
char *name;
{
	int fd;
	struct stat statbuf;
	char *databuf;
	register char *cp, *cp2;
	int len;
	int ll;

	if (!name || !*name)
		name = lastln;

	fd = openlastlines(homenod.namval, name, O_RDONLY);
	if (fd < 0)
		return -1;
	if (fstat(fd, &statbuf) < 0)
	{
		return -1;
	}
	databuf = (char *)malloc(len = statbuf.st_size);
	if ((int)databuf < 1)
		return -1;
	if (read(fd, databuf, len) != len)
	{
		CLOSE(fd);
		return -1;
	}
	CLOSE(fd);
	cp = databuf;
	ll = nlastlines-1;
	while (len > 0 && ll >= 0)
	{
		cp2 = cp;
		while (*cp2 != '\n' && len)
		{
			cp2++;
			len--;
		}
		if (cp2 != cp)
		{
			*cp2 = 0;
			lastlinesll-- = roomfor(cp);
		}
		else
			break;
		cp2++;
		len--;
		cp = cp2;
	}
	while (ll >= 0)
	{
		lastlinesll = (char *)malloc(2);
		*(lastlinesll--) = 0;
	}
	free(databuf);
	return 0;
}

countwords( line )
char *line;
{
	register int count = 0;
	register char *cp;
	register int inword = 0;

	cp = line;
	while( *cp )
	{
		if( *cp != ' ' && *cp != '\t' && !inword )
		{
			count++;
			inword = 1;
		}
		else if( *cp == ' ' || *cp == '\t' ) /* End of word */
			inword = 0;

		++cp;
	}
	return( count );
}

static char wbuf128;

char *
getword( line, posn )
char *line;
int posn;
{
	register int cw = 0;
	register char *cp, *cp1;
	int	inword = 0;

	cp = line;
	while( *cp )
	{
		if( *cp != ' ' && *cp != '\t' && !inword )
		{
			cw++;
			if( cw == posn )
				break;
			inword = 1;
		}
		else if( *cp == ' ' || *cp == '\t' ) /* End of word */
			inword = 0;

		++cp;
	}

	/* At the beginning of the word */
	cp1 = wbuf;
	if( *cp && cw == posn )
	{
		while( *cp && *cp != ' ' && *cp != '\t' )
			*cp1++ = *cp++;

		*cp1++ = ' ';
		*cp1 = (char)0;
	}
	else
		return( (char *)0 );

	return( wbuf );
}
